#!/usr/bin/env python

## Copyright 2018-2020 Stian Soiland-Reyes, The University of Manchester, UK
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

"""
Create/parse arcp (Archive and Package) URIs.

This module provides functions for creating arcp_ URIs, 
which can be used for identifying or parsing hypermedia 
files packaged in an archive like a ZIP file::

    >>> from arcp import *

    >>> arcp_random()
    'arcp://uuid,dcd6b1e8-b3a2-43c9-930b-0119cf0dc538/'

    >>> arcp_random("/foaf.ttl", fragment="me")
    'arcp://uuid,dcd6b1e8-b3a2-43c9-930b-0119cf0dc538/foaf.ttl#me'

    >>> arcp_hash(b"Hello World!", "/folder/")
    'arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/folder/'

    >>> arcp_location("http://example.com/data.zip", "/file.txt")
    'arcp://uuid,b7749d0b-0e47-5fc4-999d-f154abe68065/file.txt'

arcp URLs can be used with :mod:`urllib.parse`, 
for instance using :func:`urllib.parse.urljoin` to resolve relative references::

    >>> css = arcp.arcp_name("app.example.com", "css/style.css")
    >>> urllib.parse.urljoin(css, "../fonts/foo.woff")
    'arcp://name,app.example.com/fonts/foo.woff'


In addition this module provides functions that can be used
to parse arcp URIs into its constituent fields::

    >>> is_arcp_uri("arcp://uuid,b7749d0b-0e47-5fc4-999d-f154abe68065/file.txt")
    True

    >>> is_arcp_uri("http://example.com/t")
    False

    >>> u = parse_arcp("arcp://uuid,b7749d0b-0e47-5fc4-999d-f154abe68065/file.txt")
    ARCPSplitResult(scheme='arcp',prefix='uuid',name='b7749d0b-0e47-5fc4-999d-f154abe68065',
      uuid='b7749d0b-0e47-5fc4-999d-f154abe68065',path='/file.txt',query='',fragment='')

    >>> u.path
    '/file.txt'
    >>> u.prefix
    'uuid'
    >>> u.uuid
    UUID('b7749d0b-0e47-5fc4-999d-f154abe68065')
    >>> u.uuid.version
    5

    >>> parse_arcp("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/folder/").hash
    ('sha-256', '7f83b1657ff1fc53b92dc18148a1d65dfc2d4b1fa3d677284addd200126d9069')

The object returned from :func:`parse_arcp()` is similar to 
:class:`urllib.parse.ParseResult`, but contains additional properties 
:attr:`prefix`, :attr:`uuid`, :attr:`ni`, :attr:`hash` and :attr:`name`, 
some of which will be ``None`` depending on the arcp prefix.

The function :func:`arcp.parse.urlparse()` can be imported as an alternative 
to :func:`urllib.parse.urlparse`. If the scheme is ``arcp`` then the extra 
arcp fields like :attr:`prefix`, :attr:`uuid`, :attr:`hash` and :attr:`name` are available
as from :func:`parse_arcp`, otherwise the output is the same as from 
:func:`urllib.parse.urlparse`::

    >>> from arcp.parse import urlparse
    >>> urlparse("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/folder/soup;sads")
    ARCPParseResult(scheme='arcp',prefix='ni',
       name='sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk',
       ni='sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk',
       hash=('sha-256', '7f83b1657ff1fc53b92dc18148a1d65dfc2d4b1fa3d677284addd200126d9069',
       path='/folder/soup;sads',query='',fragment='')
    >>> urlparse("http://example.com/help?q=a")
    ParseResult(scheme='http', netloc='example.com', path='/help', params='', 
      query='q=a', fragment='')


.. _arcp: https://tools.ietf.org/html/draft-soilandreyes-arcp-03
"""

__author__      = "Stian Soiland-Reyes <https://orcid.org/0000-0001-9842-9718>"
__copyright__   = "Copyright 2018-2020 The University of Manchester"
__license__     = "Apache License, version 2.0 <https://www.apache.org/licenses/LICENSE-2.0>"

ARCP="arcp"
NI="ni"
NIH="ni"

try:
    import urllib.parse as _urlp
except:
    import urlparse as _urlp

def _register_scheme(scheme, *uses):
    """Ensure app scheme works with :func:`urllib.parse.urljoin` and friends"""    
    for u in uses:
        if not scheme in u:
            u.append(scheme)
            
_register_scheme(ARCP, _urlp.uses_relative, _urlp.uses_netloc, 
                _urlp.uses_params, _urlp.uses_fragment) # arcp://a/b;c?d#e
_register_scheme(NI, _urlp.uses_netloc) # ni://a/b
_register_scheme(NIH) # nih:a;b


# Convenience export of public functions
from .parse import is_arcp_uri, parse_arcp
from .generate import arcp_uuid, arcp_random, arcp_location, arcp_name, arcp_hash



#!/usr/bin/env python

## Copyright 2018-2020 Stian Soiland-Reyes, The University of Manchester, UK
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

"""
Generate arcp URIs with various prefixes.

As detailed in draft-soilandreyes-arcp_, the choice of 
arcp _prefix_ depends on the uniqueness constraints required
to identify the archive.

:func:`arcp_random()` can be used for a fresh arcp URI 
based on a pseudo-random generator. Use 
:func:`urllib.parse.urljoin()` to 
resolve paths within the same archive.

:func:`arcp_uuid()` can be used with a pre-made UUID instance,
for instance loaded from an archive's manifest
or generated with :func:`uuid.uuid4()`

:func:`arcp_location()` can be used to identify an archive based on
its location URL, facilitating a UUID v5 authority.

:func:`arcp_name()` can be used to identify an archive based on its
absolute DNS name or package name within an installation.

.. _draft-soilandreyes-arcp: https://tools.ietf.org/id/draft-soilandreyes-arcp-03.html
"""

__author__      = "Stian Soiland-Reyes <http://orcid.org/0000-0001-9842-9718>"
__copyright__   = "Copyright 2018 The University of Manchester"
__license__     = "Apache License, version 2.0 (https://www.apache.org/licenses/LICENSE-2.0)"

from uuid import uuid4, uuid5, UUID, NAMESPACE_URL

try:
    from urllib.parse import urlunsplit
except:
    from urlparse import urlunsplit

import re
from hashlib import sha256
from base64 import urlsafe_b64encode, urlsafe_b64decode

SCHEME="arcp"

def _reg_name_regex():
    """Compile regular expression for RFC3986_ reg-name production

    _RFC3986: https://www.ietf.org/rfc/rfc3986
    """
    # unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
    unreserved = r"[A-Za-z0-9-._~]"

    # pct-encoded = "%" HEXDIG HEXDIG
    pct_encoded = r"%[0-9A-Fa-f][0-9A-Fa-f]"
    
    # "!" / "$" / "&" / "'" / "(" / ")"
    #  / "*" / "+" / "," / ";" / "="
    sub_delims = r"[!$&'()*+,;=]"

    # reg-name    = *( unreserved / pct-encoded / sub-delims )    
    reg_name = r"^(" + unreserved + r"|" + pct_encoded + sub_delims + r")*$"
    return re.compile(reg_name)
_REG_NAME = _reg_name_regex()

def arcp_uuid(uuid, path="/", query=None, fragment=None):
    """Generate an arcp URI for the given uuid.

    Parameters:
      - uuid -- a uuid string or UUID instance identifying the archive, e.g. ``58ca7fa6-be2f-48e4-8b69-e63fb0d929fe``
      - path -- Optional path within archive.
      - query -- Optional query component.
      - fragment -- Optional fragment component.
    """
    if not isinstance(uuid, UUID):
        # ensure valid UUID
        uuid = UUID(uuid)

    # TODO: Ensure valid path?    
    path = path or ""
    authority = "uuid,%s" % uuid
    s = (SCHEME, authority, path, query, fragment)
    return urlunsplit(s)

def arcp_random(path="/", query=None, fragment=None, uuid=None):
    """Generate an arcp URI using a random uuid.

    Parameters:
      - path -- Optional path within archive.
      - query -- Optional query component.
      - fragment -- Optional fragment component.
      - uuid -- optional UUID v4 string or UUID instance
    """
    if uuid is None:
        uuid = uuid4()
    elif not isinstance(uuid, UUID):
        # ensure valid UUID
        uuid = UUID(uuid)
    if not uuid.version == 4:
        raise Exception("UUID is not v4" % uuid)
    return arcp_uuid(uuid, path=path, query=query, fragment=fragment)

def arcp_location(location, path="/", query=None, fragment=None, namespace=NAMESPACE_URL):
    """Generate an arcp URI for a given archive location.

    Parameters:
      - location: URL or location of archive, e.g. ``http://example.com/data.zip``
      - path -- Optional path within archive.
      - query -- Optional query component.
      - fragment -- Optional fragment component.
      - namespace -- optional namespace UUID for non-URL location.
    """
    # TODO: Ensure location is valid url if NAMESPACE_URL?
    uuid = uuid5(namespace, location)
    return arcp_uuid(uuid, path=path, query=query, fragment=fragment)
    
def arcp_name(name, path="/", query=None, fragment=None):
    """Generate an arcp URI for a given archive name.

    Parameters:
      - name -- Absolute DNS or package name, e.g. ``app.example.com``
      - path -- Optional path within archive.
      - query -- Optional query component.
      - fragment -- Optional fragment component.
      - namespace -- optional namespace UUID for non-URL location.
    """
    if not _REG_NAME.match(name):
        raise Exception("Invalid name: %s" % name)
    authority = "name," + name
    s = (SCHEME, authority, path, query, fragment)
    return urlunsplit(s)

def arcp_hash(bytes=b"", path="/", query=None, fragment=None, hash=None):
    """Generate an arcp URI for a given archive hash checksum.

    Parameters:
      - bytes -- Optional bytes of archive to checksum
      - path -- Optional path within archive.
      - query -- Optional query component.
      - fragment -- Optional fragment component.
      - hash -- Optional hash instance from :func:`hashlib.sha256()`
    
    Either ``bytes`` or ``hash`` must be provided. 
    The ``hash`` parameter can be provided to avoid representing 
    the whole archive bytes in memory.
    """
    if hash is None:
        hash = sha256()
    elif hash.name != "sha256":
        # TODO: Map Python's hash-names to RFC6920
        raise Exception("hash method %s unsupported, try sha256" % hash.name)
    hashmethod = "sha-256"

    # Tip: if bytes == b"" then provided hash param is unchanged
    hash.update(bytes)

    # RFC6920-style hash encoding
    digestB64 = urlsafe_b64encode(hash.digest())
    digestB64 = digestB64.decode("ascii").strip("=")
    authority = "ni,%s;%s" % (hashmethod, digestB64)
    s = (SCHEME, authority, path, query, fragment)
    return urlunsplit(s)




#!/usr/bin/env python

## Copyright 2018 Stian Soiland-Reyes, The University of Manchester, UK
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.


import unittest

from arcp import parse

class TestIsArcpURI(unittest.TestCase):
    """Test is_arcp_uri()"""

    def test_arcp_uri(self):
        self.assertTrue(parse.is_arcp_uri(
            "arcp://uuid,ecba06ed-472e-46d4-8ab8-9570e40e0b8c/"))

    def test_not_arcp_uri(self):
        self.assertFalse(parse.is_arcp_uri(
            "http://example.com/"))

    def test_arcp_uri_authority_fallback(self):
        self.assertTrue(parse.is_arcp_uri(
            "arcp://example.com"))
        self.assertTrue(parse.is_arcp_uri(
            "arcp://x-unknown,abc"))


class TestParse(unittest.TestCase):
    """Test parse_arcp()"""

    def _test_tuple(self, t):
        (scheme,netloc,path,params,query,fragment) = t
        self.assertEqual(t.scheme, scheme)
        self.assertEqual(t.netloc, netloc)
        self.assertEqual(t.path, path)
        self.assertEqual(t.params, params)
        self.assertEqual(t.query, query)
        self.assertEqual(t.fragment, fragment)

    def test_parse(self):
        t = parse.parse_arcp("arcp://uuid,ecba06ed-472e-46d4-8ab8-9570e40e0b8c/")
        self._test_tuple(t)
        self.assertEqual("arcp", t.scheme)
        self.assertEqual("uuid,ecba06ed-472e-46d4-8ab8-9570e40e0b8c", t.netloc)
        self.assertEqual("/", t.path)
        self.assertEqual("", t.params)
        self.assertEqual("", t.query)
        self.assertEqual("", t.fragment)

    def test_parse_path_query_fragment(self):
        t = parse.parse_arcp("arcp://uuid,ecba06ed-472e-46d4-8ab8-9570e40e0b8c/file;p=1?q=a#frag")
        self._test_tuple(t)
        self.assertEqual("arcp", t.scheme)
        self.assertEqual("uuid,ecba06ed-472e-46d4-8ab8-9570e40e0b8c", t.netloc)
        self.assertEqual("/file", t.path)
        self.assertEqual("p=1", t.params)
        self.assertEqual("q=a", t.query)
        self.assertEqual("frag", t.fragment)

    def test_parse_authority_fallback(self):
        t = parse.parse_arcp("arcp://example.com/")
        self._test_tuple(t)
        self.assertEqual("arcp", t.scheme)
        self.assertEqual("example.com", t.netloc)
        self.assertEqual("/", t.path)
        self.assertEqual("", t.params)
        self.assertEqual("", t.query)
        self.assertEqual("", t.fragment)

    def test_parse_empty_authority(self):
        t = parse.parse_arcp("arcp:///")
        self._test_tuple(t)
        self.assertEqual("arcp", t.scheme)
        self.assertEqual("", t.netloc)
        self.assertEqual("/", t.path)
        self.assertEqual("", t.params)
        self.assertEqual("", t.query)
        self.assertEqual("", t.fragment)


    def test_parseFails(self):
        with self.assertRaises(Exception):
            parse.parse_arcp("http://example.com/")

    def test_parse_prefix(self):
        self.assertEqual("uuid",
             parse.parse_arcp("arcp://uuid,ecba06ed-472e-46d4-8ab8-9570e40e0b8c/").prefix)
        self.assertEqual("ni",
             parse.parse_arcp("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/").prefix)
        self.assertEqual("name",
             parse.parse_arcp("arcp://name,example.com/").prefix)

        # authority fall-back
        self.assertIsNone(
             parse.parse_arcp("arcp://example.com/").prefix)
        # silly..but valid by specification
        self.assertEqual("",
             parse.parse_arcp("arcp://,example.com/").prefix)
        # Unknown prefixes are also picked up (should they?)
        self.assertEqual("x-unknown",
             parse.parse_arcp("arcp://x-unknown,abc").prefix)

    def test_parse_name(self):
        self.assertEqual("example.com",
             parse.parse_arcp("arcp://name,example.com/").name)
        # but other "names" are also supported:
        self.assertEqual("ecba06ed-472e-46d4-8ab8-9570e40e0b8c",
             parse.parse_arcp("arcp://uuid,ecba06ed-472e-46d4-8ab8-9570e40e0b8c/").name)
        self.assertEqual("sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk",
             parse.parse_arcp("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/").name)
        # authority fall-back
        self.assertEqual("example.com",
             parse.parse_arcp("arcp://example.com/").name)
        # unlikely, but valid by specification
        self.assertEqual("",
             parse.parse_arcp("arcp:///").name)
        # name from an unknown prefix
        self.assertEqual("abc",
             parse.parse_arcp("arcp://x-unknown,abc").name)


    def test_parse_uuid(self):
        self.assertEqual("ecba06ed472e46d48ab89570e40e0b8c",
             parse.parse_arcp("arcp://uuid,ecba06ed-472e-46d4-8ab8-9570e40e0b8c/").uuid.hex)
        self.assertIsNone(
            parse.parse_arcp("arcp://name,ecba06ed-472e-46d4-8ab8-9570e40e0b8c/").uuid)
        self.assertIsNone(
            parse.parse_arcp("arcp://ecba06ed-472e-46d4-8ab8-9570e40e0b8c/").uuid)
        
    def test_parse_uuid_fails(self):
        with self.assertRaises(Exception):
            parse.parse_arcp("arcp://uuid,ecba06ed-WRONG/").uuid

    def test_parse_ni(self):
        self.assertEqual("sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk",
             parse.parse_arcp("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/").ni)
        self.assertIsNone(
            parse.parse_arcp("arcp://name,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/").ni)
        self.assertIsNone(
            parse.parse_arcp("arcp://sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/").ni)

    def test_parse_ni_invalid_fails(self):
        # Below example is invalid as alg-val string 
        # does not contain ;
        with self.assertRaises(Exception):
             parse.parse_arcp("arcp://ni,sha-256/").ni_uri()
        with self.assertRaises(Exception):
             parse.parse_arcp("arcp://ni,sha-256/").nih_uri()
        with self.assertRaises(Exception):
             parse.parse_arcp("arcp://ni,sha-256/").hash()
        with self.assertRaises(Exception):
             parse.parse_arcp("arcp://ni,sha-256/").ni_well_known()
        with self.assertRaises(Exception):
             parse.parse_arcp("arcp://ni,sha-256/").ni

    def test_parse_ni_uri(self):
        self.assertEqual("ni:///sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk",
             parse.parse_arcp("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .ni_uri())
        self.assertIsNone(
            parse.parse_arcp("arcp://name,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .ni_uri())
        self.assertIsNone(
            parse.parse_arcp("arcp://sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .ni_uri())
        self.assertEqual("ni://example.com/sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk",
             parse.parse_arcp("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .ni_uri("example.com"))

    def test_parse_nih_uri(self):
        self.assertEqual("nih:sha-256-120;532690-57e12f-e2b74b-a07c89-2560a2;f",
             parse.parse_arcp("arcp://ni,sha-256-120;UyaQV-Ev4rdLoHyJJWCi/")
                .nih_uri())
        self.assertIsNone(
            parse.parse_arcp("arcp://name,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .nih_uri())
        self.assertIsNone(
            parse.parse_arcp("arcp://sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .nih_uri())

    def test_parse_ni_well_known(self):
        self.assertEqual("/.well-known/ni/sha-256/f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk",
             parse.parse_arcp("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .ni_well_known())
        self.assertIsNone(
            parse.parse_arcp("arcp://name,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .ni_well_known())
        self.assertIsNone(
            parse.parse_arcp("arcp://sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .ni_well_known())
        self.assertEqual("http://example.com/.well-known/ni/sha-256/f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk",
             parse.parse_arcp("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .ni_well_known("http://example.com/"))

    def test_parse_hash(self):
        # sha256 of "Hello World!" in ascii
        self.assertEqual(("sha-256", "7f83b1657ff1fc53b92dc18148a1d65dfc2d4b1fa3d677284addd200126d9069"),
             parse.parse_arcp("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .hash)
        self.assertIsNone(
            parse.parse_arcp("arcp://name,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .hash)
        self.assertIsNone(
            parse.parse_arcp("arcp://sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/")
                .hash)
        # md5 of "Hello World!" in ascii
        self.assertEqual(("md5", "ed076287532e86365e841e92bfc50d8c"),
             parse.parse_arcp("arcp://ni,md5;7Qdih1MuhjZehB6Sv8UNjA/")
                .hash)

    def test_parse_repr_ni(self):
        u = parse.parse_arcp("arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/file?q=a#frag")
        r = repr(u)
        self.assertIn("scheme='arcp'", r)
        self.assertIn("prefix='ni'", r)
        self.assertIn("name='sha-256;f4OxZX", r)#..
        self.assertIn("path='/file'", r)
        self.assertIn("query='q=a'", r)
        self.assertIn("fragment='frag'", r)
        self.assertIn("ni='sha-256;f4OxZX", r)#..
        self.assertIn("hash=('sha-256', '7f83b16", r)#..
        self.assertNotIn("uuid=", r)

    def test_parse_repr_uuid(self):
        u = parse.parse_arcp("arcp://uuid,32a423d6-52ab-47e3-a9cd-54f418a48571/")
        r = repr(u)
        self.assertIn("scheme='arcp'", r)
        self.assertIn("prefix='uuid'", r)
        self.assertIn("name='32a423d6-", r)#..
        self.assertIn("path='/'", r)
        self.assertIn("uuid=32a423d6-", r)# ..
        self.assertIn("query=''", r)
        self.assertIn("fragment=''", r)
        self.assertNotIn("ni=", r)
        self.assertNotIn("hash=", r)#..

    def test_parse_str_ni(self):
        uri = "arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/file?q=a#frag"
        u = parse.parse_arcp(uri)
        self.assertEqual(uri, str(u))
        

class URLParse(unittest.TestCase):
    """Test urlparse()"""
    def test_urlparse(self):
        self.assertEqual("name",
            parse.urlparse("arcp://name,example.com/").prefix)
        self.assertEqual("http",
            parse.urlparse("http://example.com/").scheme)

class NIH_CheckDigit(unittest.TestCase):
    """Test _nih_checkdigit() using RFC6920 examples"""
    def test_checkdigit(self):
        self.assertEqual("f", parse._nih_checkdigit(
          "5326-9057-e12f-e2b7-4ba0-7c89-2560-a2".replace("-", "")))
        self.assertEqual("b", parse._nih_checkdigit("53269057"))                
        self.assertEqual("0", parse._nih_checkdigit("b053269057"))
        self.assertEqual("d", parse._nih_checkdigit("acefeed"))
        self.assertEqual("0", parse._nih_checkdigit("dacefeed"))
        self.assertEqual("4", parse._nih_checkdigit("123456789abcdef"))
        self.assertEqual("0", parse._nih_checkdigit("4123456789abcdef"))
        # Consistency check -- if we add $digit (or $digit0 for even-length) 
        # in front, the new sum should be 0

class NIH_Segmented(unittest.TestCase):
    """Test _nih_segmented()"""
    def test_segment(self):
        self.assertEqual("532690-57e12f-e2b74b-a07c89-2560a2", 
            parse._nih_segmented("53269057e12fe2b74ba07c892560a2"))
        self.assertEqual("5326-9057-e12f-e2b7-4ba0-7c89-2560-a2", 
            parse._nih_segmented("53269057e12fe2b74ba07c892560a2", 4))
        self.assertEqual("532690-5", 
            parse._nih_segmented("5326905"))
        self.assertEqual("532690", 
            parse._nih_segmented("532690"))
        self.assertEqual("53269", 
            parse._nih_segmented("53269"))
        self.assertEqual("", 
            parse._nih_segmented(""))



#!/usr/bin/env python

## Copyright 2018 Stian Soiland-Reyes, The University of Manchester, UK
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

import unittest
from uuid import UUID, RFC_4122, NAMESPACE_OID
import re

from arcp import generate
from hashlib import sha256, md5

# Some test data
TEST_UUID_v1 = UUID("dbc0802a-0682-11e8-9895-b8ca3ad10ac0")
TEST_UUID_v4 = UUID("8c36d39a-18be-4aa8-b1ce-fef330b00a28")
OID = "1.3.6.1.4.1.13661"

uuid_re = re.compile(r"^[0-9a-fA-F]{8}-([0-9a-fA-F]{4}-){3}[0-9a-fA-F]{12}$")


class UUIDTest(unittest.TestCase):
    """Test arcp_uuid()"""
    def testUUID(self):
        self.assertEqual("arcp://uuid,dbc0802a-0682-11e8-9895-b8ca3ad10ac0/",
            generate.arcp_uuid(TEST_UUID_v1))
        # UUID version should not matter
        self.assertEqual("arcp://uuid,8c36d39a-18be-4aa8-b1ce-fef330b00a28/",
            generate.arcp_uuid(TEST_UUID_v4))            
    def testUUIDPath(self):
        self.assertEqual("arcp://uuid,dbc0802a-0682-11e8-9895-b8ca3ad10ac0/folder/file.txt",
            generate.arcp_uuid(TEST_UUID_v1, "/folder/file.txt"))
    def testUUIDPathQuery(self):
        self.assertEqual("arcp://uuid,dbc0802a-0682-11e8-9895-b8ca3ad10ac0/folder/file.txt?q=s",
            generate.arcp_uuid(TEST_UUID_v1, "/folder/file.txt", "q=s"))
    def testUUIDPathQueryFrag(self):
        self.assertEqual("arcp://uuid,dbc0802a-0682-11e8-9895-b8ca3ad10ac0/folder/file.txt?q=s#frag",
            generate.arcp_uuid(TEST_UUID_v1, "/folder/file.txt", "q=s", "frag"))
    def testUUIDQuery(self):
        self.assertEqual("arcp://uuid,dbc0802a-0682-11e8-9895-b8ca3ad10ac0/?a=b&c=d",
            generate.arcp_uuid(TEST_UUID_v1, query="a=b&c=d"))
    def testUUIDFrag(self):
        self.assertEqual("arcp://uuid,dbc0802a-0682-11e8-9895-b8ca3ad10ac0/#frag",
            generate.arcp_uuid(TEST_UUID_v1, fragment="frag"))
    def testUUIDstr(self):
        self.assertEqual("arcp://uuid,dbc0802a-0682-11e8-9895-b8ca3ad10ac0/",
            generate.arcp_uuid("dbc0802a-0682-11e8-9895-b8ca3ad10ac0"))
    def testUUIDstrFromUpperCase(self):
        self.assertEqual("arcp://uuid,dbc0802a-0682-11e8-9895-b8ca3ad10ac0/",
            generate.arcp_uuid("dbc0802a-0682-11e8-9895-b8ca3ad10ac0"))    
    def testUUIDstrPathQueryFrag(self):
        self.assertEqual("arcp://uuid,dbc0802a-0682-11e8-9895-b8ca3ad10ac0/folder/file.txt?q=s#frag",
            generate.arcp_uuid("dbc0802a-0682-11e8-9895-b8ca3ad10ac0", 
                               "/folder/file.txt", "q=s", "frag"))
    def testUUIDstrInvalidUUID(self):
        with self.assertRaises(Exception):
            # Too short
            generate.arcp_uuid("5da78af6")
        with self.assertRaises(Exception):
            # empty
            generate.arcp_uuid("")


class RandomTest(unittest.TestCase):
    """Test arcp_random(), with implicit or explicit UUID"""
    def testRandom(self):
        u = generate.arcp_random()
        u2 = generate.arcp_random()
        # always fresh
        self.assertNotEqual(u, u2) 
        self.assertTrue(u.startswith("arcp://uuid,"))
        self.assertTrue(u.endswith("/"))
    def testRandomValidUUID(self):
        u = generate.arcp_random()
        # Extract UUID to ensure it is v4
        uuidStr = u.replace("arcp://uuid,", "").strip("/")
        # Ensure RFC4122 compliance
        self.assertIsNotNone(uuid_re.match(uuidStr))
        # ensure lower-case in output
        self.assertEqual(uuidStr, uuidStr.lower())
        # must be valid UUID
        uuid = UUID(uuidStr)
        # must be RFC_4122 variant, version 4 (random)
        self.assertEqual(RFC_4122, uuid.variant)
        self.assertEqual(4, uuid.version)
        
    def testRandomPath(self):
        u = generate.arcp_random("/folder/file.txt")
        self.assertTrue(u.startswith("arcp://uuid,"))
        self.assertTrue(u.endswith("/folder/file.txt"))
        self.assertEqual(
            # initial / in path is optional
            len(generate.arcp_random("file.txt")), 
            len(generate.arcp_random("/file.txt")))
    def testRandomQuery(self):
        u = generate.arcp_random(query="q=a")
        self.assertTrue(u.startswith("arcp://uuid,"))
        self.assertTrue(u.endswith("/?q=a"))
    def testRandomFrag(self):
        u = generate.arcp_random(fragment="frag")
        self.assertTrue(u.startswith("arcp://uuid,"))
        self.assertTrue(u.endswith("/#frag"))
    def testRandomPathQueryFrag(self):
        u = generate.arcp_random("/folder/file.txt", "a=b&c=d", "frag")
        self.assertTrue(u.startswith("arcp://uuid,"))
        self.assertTrue(u.endswith("/folder/file.txt?a=b&c=d#frag"))

    # Now test providing a fixed UUID        
    def testUUID(self):
        self.assertEqual("arcp://uuid,8c36d39a-18be-4aa8-b1ce-fef330b00a28/",
            generate.arcp_random(uuid=TEST_UUID_v4))
        # UUID version must be 4
        with self.assertRaises(Exception):
            generate.arcp_random(uuid=TEST_UUID_v1)
    def testUUIDPath(self):
        self.assertEqual("arcp://uuid,8c36d39a-18be-4aa8-b1ce-fef330b00a28/folder/file.txt",
            generate.arcp_random("/folder/file.txt", uuid=TEST_UUID_v4))
    def testUUIDstr(self):
        self.assertEqual("arcp://uuid,8c36d39a-18be-4aa8-b1ce-fef330b00a28/",
            generate.arcp_random(uuid="8c36d39a-18be-4aa8-b1ce-fef330b00a28"))
        with self.assertRaises(Exception):
            # UUID version must be 4
            generate.arcp_random(uuid="dbc0802a-0682-11e8-9895-b8ca3ad10ac0")
            
    def testUUIDstrPathQueryFrag(self):
        self.assertEqual("arcp://uuid,8c36d39a-18be-4aa8-b1ce-fef330b00a28/folder/file.txt?q=s#frag",
            generate.arcp_random("/folder/file.txt", "q=s", "frag", 
                uuid="8c36d39a-18be-4aa8-b1ce-fef330b00a28"))
    def testUUIDstrInvalidUUID(self):
        with self.assertRaises(Exception):
            # Too short
            generate.arcp_random(uuid="5da78af6")
        with self.assertRaises(Exception):
            # empty
            generate.arcp_random(uuid="")      


class LocationTest(unittest.TestCase):
    """Test arcp_location()"""
    def testExampleZip(self):
        # URL and expected UUID as in
        # https://tools.ietf.org/id/draft-soilandreyes-arcp-03.html#rfc.appendix.A.2
        self.assertEqual("arcp://uuid,b7749d0b-0e47-5fc4-999d-f154abe68065/", 
            generate.arcp_location("http://example.com/data.zip"))
        self.assertEqual("arcp://uuid,b7749d0b-0e47-5fc4-999d-f154abe68065/pics/", 
            generate.arcp_location("http://example.com/data.zip", "/pics/"))
        self.assertEqual("arcp://uuid,b7749d0b-0e47-5fc4-999d-f154abe68065/pics/flower.jpeg", 
            generate.arcp_location("http://example.com/data.zip", "/pics/flower.jpeg"))

    def testExampleZipUUIDValid(self):
        u = generate.arcp_location("http://example.com/")
        # Extract UUID to ensure it is v4
        uuidStr = u.replace("arcp://uuid,", "").strip("/")
        # Ensure RFC4122 compliance
        self.assertIsNotNone(re.match(uuid_re, uuidStr))
        # ensure lower-case in output
        self.assertEqual(uuidStr, uuidStr.lower())
        # must be valid UUID
        uuid = UUID(uuidStr)
        # must be RFC_4122 variant, version 5 (name sha1)
        self.assertEqual(RFC_4122, uuid.variant)
        self.assertEqual(5, uuid.version)

    def testLocationOtherNamespace(self):
        u = generate.arcp_location(OID, namespace=NAMESPACE_OID, path="/example")
        self.assertEqual("arcp://uuid,215aa48f-233f-507f-8484-3eb5d6e23e9d/example", u)

    def testLocationNamespaceUUIDstrPathQueryFrag(self):
        self.assertEqual("arcp://uuid,215aa48f-233f-507f-8484-3eb5d6e23e9d/folder/file.txt?q=s#frag",
            generate.arcp_location(OID, "/folder/file.txt", "q=s", "frag", 
                namespace=NAMESPACE_OID))


class NameTest(unittest.TestCase):
    """Test arcp_name()"""
    def testExampleName(self):
        # Adapted from 
        # https://tools.ietf.org/id/draft-soilandreyes-arcp-02.html#rfc.appendix.A.7
        self.assertEqual("arcp://name,gallery.example.org/",
            generate.arcp_name("gallery.example.org"))
        new_photos = generate.arcp_name("gallery.example.org", "/photos/", "New")
        
        self.assertEqual("arcp://name,gallery.example.org/photos/?New",
            new_photos)
        photo = generate.arcp_name("gallery.example.org", "/photos/137")
        self.assertEqual("arcp://name,gallery.example.org/photos/137", photo)

        template = "arcp://name,messaging.example.com/share;{*uri};{*redirect}"
        u = template.replace("{*uri}", photo).replace("{*redirect}", new_photos)

        # TODO: Should arcp_name support 
        path =  "/share;%s;%s" % (photo, new_photos)
        self.assertEqual(u, 
            generate.arcp_name("messaging.example.com", path))

    def testNamePath(self):
        self.assertEqual("arcp://name,app.example.org/msgs/1",
            generate.arcp_name("app.example.org", "/msgs/1"))

    def testNameQuery(self):
        self.assertEqual("arcp://name,app.example.org/?q=a",
            generate.arcp_name("app.example.org", query="q=a"))

    def testNameHash(self):
        self.assertEqual("arcp://name,app.example.org/#frag",
            generate.arcp_name("app.example.org", fragment="frag"))

    def testNamePathQueryHash(self):
        self.assertEqual("arcp://name,app.example.org/msgs/1?a=b&c=d#frag",
            generate.arcp_name("app.example.org", "/msgs/1", "a=b&c=d", "frag"))

    def testInvalidName(self):
        with self.assertRaises(Exception):
            generate.arcp_name("example com")

    def testEmptyName(self):
        # empty name is valid by the spec (TODO: should it be?)
        self.assertEqual("arcp://name,/",
            generate.arcp_name(""))
            

# Example from https://tools.ietf.org/html/rfc6920#section-8.1
BYTES = "Hello World!".encode("ASCII")
# echo -n "Hello World!" | sha256sum
HASH = "7f83b1657ff1fc53b92dc18148a1d65dfc2d4b1fa3d677284addd200126d9069"
NI = "ni:///sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk"
ARCP = "arcp://ni,sha-256;f4OxZX_x_FO5LcGBSKHWXfwtSx-j1ncoSt3SABJtkGk/"

class HashTest(unittest.TestCase):
    """Test arcp_hash()"""
    def testHash(self):
        # Using "Hello World!" example
        # instead of a real archive file
        self.assertEqual(ARCP, generate.arcp_hash(BYTES))
        
    def testHashPath(self):
        self.assertEqual(ARCP + "bin/evil", 
            generate.arcp_hash(BYTES, "/bin/evil"))

    def testHashQuery(self):
        self.assertEqual(ARCP + "?q=a", 
            generate.arcp_hash(BYTES, query="q=a"))
    
    def testHashFragment(self):
        self.assertEqual(ARCP + "#frag", 
            generate.arcp_hash(BYTES, fragment="frag"))

    def testHashPathQueryFragment(self):
        self.assertEqual(ARCP + "bin/evil?a=b&c=d#frag", 
            generate.arcp_hash(BYTES, "/bin/evil", "a=b&c=d", "frag"))

    def testHashHash(self):
        h = sha256(BYTES)
        self.assertEqual(ARCP, generate.arcp_hash(hash=h))

    def testHashPathEmptyHash(self):
        h = sha256()
        self.assertEqual(ARCP + "bin/evil",
            generate.arcp_hash(BYTES, "bin/evil", hash=h))

    def testHashPathPremadeHash(self):
        h = sha256(BYTES)
        self.assertEqual(ARCP + "bin/evil",
            generate.arcp_hash(path="bin/evil", hash=h))

    def testHashWrongHash(self):
        h = md5(BYTES)
        with self.assertRaises(Exception):
            generate.arcp_hash(hash=h)




#!/usr/bin/env python

## Copyright 2018-2020 Stian Soiland-Reyes, The University of Manchester, UK
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

__author__      = "Stian Soiland-Reyes <https://orcid.org/0000-0001-9842-9718>"
__copyright__   = "Copyright 2018-2020 The University of Manchester"
__license__     = "Apache License, version 2.0 (https://www.apache.org/licenses/LICENSE-2.0)"

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
  long_description = f.read()


setup(
  name = 'arcp',
  packages = find_packages(exclude=['contrib', 'docs', 'tests']), # Required
  version = '0.2.1',
  description = 'arcp (Archive and Package) URI parser and generator',
  long_description=long_description,
  author = 'Stian Soiland-Reyes',
  author_email = 'stain@apache.org',
  
  # https://www.apache.org/licenses/LICENSE-2.0
  license = "Apache License, Version 2.0",
  url = 'https://arcp.readthedocs.io/',
  download_url = 'https://github.com/stain/arcp-py/archive/0.1.0.tar.gz',
  keywords = "arcp uri url iri archive package",
  
  install_requires=[],
  classifiers=[
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    'Development Status :: 3 - Alpha',

    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'Topic :: Software Development :: Libraries :: Python Modules',
     # 'License :: OSI Approved :: Apache Software License',
     # https://github.com/pypa/pypi-legacy/issues/564
    #'License :: OSI Approved',
    # 'License :: OSI Approved :: Apache License, Version 2.0 (Apache-2.0)',  
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Topic :: Internet',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: System :: Archiving',
    'Topic :: System :: Archiving :: Packaging',
],
  
)



#!/usr/bin/env python
## Copyright 2018-2020 Stian Soiland-Reyes, The University of Manchester, UK
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

"""
Parse arcp URIs.

Use is_arcp_uri() to detect of an URI string is using the 
arcp: URI scheme, in which case parse_arcp() can be used
to split it into its components.

The urlparse() function can be used as a replacement for
urllib.parse.urlparse() - supporting any URIs. If the URI is 
using the arcp: URI scheme, additional components are available
as from parse_arcp().
"""
__author__      = "Stian Soiland-Reyes <https://orcid.org/0000-0001-9842-9718>"
__copyright__   = "Copyright 2018-2020 The University of Manchester"
__license__     = "Apache License, version 2.0 (https://www.apache.org/licenses/LICENSE-2.0)"

from uuid import UUID, NAMESPACE_URL

try:
    import urllib.parse as urlp
except:
    import urlparse as urlp

from base64 import urlsafe_b64decode
from binascii import hexlify
import re

SCHEME="arcp"

def is_arcp_uri(uri):
    """Return True if the uri string uses the arcp scheme, otherwise False.
    """

    # tip: urllib will do lowercase for us
    return urlp.urlparse(uri).scheme == SCHEME

def parse_arcp(uri):
    """Parse an arcp URI string into its constituent parts.

    The returned object is similar to ``urllib.parse.urlparse()``
    in that it is a tuple of 
    ``(scheme,netloc,path,params,query,fragment)``
    with equally named properties, but it also adds
    properties for arcp fields:
    
    - prefix -- arcp authority prefix, e.g. "uuid", "ni" or "name", or None if prefix is missing
    - name -- arcp authority without prefix, e.g. "a4889890-a50a-4f14-b4e7-5fd83683a2b5" or "example.com"
    - uuid -- a ``uuid.UUID`` object if prefix is "uuid", otherwise None
    - ni -- the arcp alg-val value according to RFC6920 if prefix is "ni", otherwise None
    - hash -- the hash method and hash as a hexstring if prefix is "ni", otherwise None
    """

    return ARCPParseResult(*urlp.urlparse(uri))

def urlparse(uri):
    """Parse any URI string into constituent parts.

    The returned object is similar to 
    :func:`urllib.parse.urlparse()`
    in that it is a tuple of 
    ``(scheme,netloc,path,params,query,fragment)``
    with equally named properties, but if the 
    URI scheme is "arcp" this also adds
    arcp properties as in :func:`parse_arcp()`.
    """
    u = urlp.urlparse(uri)    
    if (u.scheme == SCHEME):
        return ARCPParseResult(*u)
    else:
        return u

class ARCPParseResult(urlp.ParseResult):
    """Result of parsing an arcp URI.

    This class does not detect if the arcp URI was valid
    according to the specification.

    This class extends :class:`urlllib.parse.ParseResult`
    adding arcp properties, some of which may be `None`.
    """
    __slots__ = ()

    def __init__(self, *args):
        if self.scheme != SCHEME:
            raise Exception("uri has scheme %s, expected %s" % 
                            (self.scheme, SCHEME))

    def _host_split(self):
        """Return (prefix,name) if authority has "," - 
        otherwise (None, authority).
        """
        if self.netloc and "," in self.netloc:
            return self.netloc.split(",", 1)
        else:
            return (None, self.netloc)

    @property
    def prefix(self):
        """The arcp prefix, e.g. "uuid", "ni", "name" or None if no prefix was present.
        """
        (prefix,name) = self._host_split()
        return prefix

    @property
    def name(self):
        """The URI's authority without arcp prefix.
        """
        (prefix,name) = self._host_split()
        return name
    
    @property
    def uuid(self):
        """The arcp UUID if the prefix is "uuid", otherwise None."""
        if self.prefix != "uuid":
            return None
        return UUID(self.name)
    
    @property
    def ni(self):
        """The arcp ni string if the prefix is "ni", otherwise None."""
        if self.prefix != "ni":
            return None
        if not _ALG_VAL.match(self.name):
            raise Exception("Invalid alg-val for ni, prefix: %s" % self.netloc)
        return self.name
    
    def ni_uri(self, authority=""):
        """The ni URI (RFC6920_) if the prefix is "ni", otherwise None.
        
        If the ``authority`` parameter is provided, 
        it will be used in the returned URI.

        .. _RFC6920: https://tools.ietf.org/search/rfc6920
        """
        ni = self.ni
        if ni is None:
            return None
        s = ("ni", authority, ni, None, None)
        return urlp.urlunsplit(s)


    def nih_uri(self):
        """The nih URI (RFC6920_) if the prefix is "ni", otherwise None.
        
        .. _RFC6920: https://tools.ietf.org/search/rfc6920
        """
        h = self.hash
        if h is None:
            return None
        (hash_method, hash_hex) = h
        
        segmented = _nih_segmented(hash_hex)
        checkdigit = _nih_checkdigit(hash_hex)

        path = "%s;%s;%s" % (hash_method, segmented, checkdigit)
        s = ("nih", None, path, None, None)
        return urlp.urlunsplit(s)
    
    def ni_well_known(self, base=""):
        """The ni .well-known URI (RFC5785_) if the prefix is 
        "ni", otherwise None.

        The parameter ``base``, if provided, should be an absolute URI like 
        ``"http://example.com/"`` - a relative URI is returned otherwise.

        .. _RFC5785: https://tools.ietf.org/html/rfc5785
        """
        (method, hash_b64) = self._ni_split()        
        if method is None:
            return None
        
        # .well-known is always at / (RFC5785)
        path = "/.well-known/ni/%s/%s" % (method, hash_b64)
        return urlp.urljoin(base, path)

    def _ni_split(self):
        """Split self.ni:
        """
        ni = self.ni
        if ni is None:
            return (None,None)
        # Already checked by self.ni regex
        #if not ";" in ni:
        #    raise Exception("invalid ni hash: %s" % ni)
        (method, hash_b64) = ni.split(";", 1)
        return (method, hash_b64)
        

    @property
    def hash(self):
        """A tuple (hash_method,hash_hex) if the prefix is "ni", 
        otherwise None.
        """
        (method, hash_b64) = self._ni_split()
        if method is None:
            return None
        # re-instate padding as urlsafe_base64decode is strict
        missing_padding = 4 - (len(hash_b64) % 4)
        hash_b64 += "=" * missing_padding
        hash_bytes = urlsafe_b64decode(hash_b64)
        hash_hex = hexlify(hash_bytes).decode("ascii")
        return (method.lower(), hash_hex)
    
    def __repr__(self):
        props = ["scheme='arcp'"]
        props += ["prefix='%s'" % self.prefix or ""]
        props += ["name='%s'" % self.name or ""]

        if self.uuid is not None:
            props += ["uuid=%s" % self.uuid]
        if self.ni is not None:
            props += ["ni='%s'" % self.ni]
            # Avoid Exception in __repr__
            if ";" in self.ni:
                props += ["hash=('%s', '%s'" % self.hash]

        # Traditional URI properties
        props += ["path='%s'" % self.path or ""]
        props += ["query='%s'" % self.query or ""]
        props += ["fragment='%s'" % self.fragment or ""]
        return "ARCPParseResult(%s)" % ",".join(props)

    def __str__(self):
        return self.geturl()

def _alg_val_regex():
    """Compile regular expression for RFC6920_ alg-val production

    .. _RFC6920: https://www.ietf.org/rfc/rfc6920
    """
    # unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
    unreserved = r"[A-Za-z0-9-._~]"
    # alg            = 1*unreserved
    alg = r"(" + unreserved + r"+)"
    # val            = 1*unreserved
    val = r"(" + unreserved + r"+)"
    # alg-val        = alg ";" val
    alg_val = r"^" + alg + ";" + val + r"$"
    return re.compile(alg_val)
_ALG_VAL = _alg_val_regex()

def _nih_segmented(h, grouping=6):
    """Segment hex-hash with dashes in nih style RFC6920_

    >>> _nih_segmented("0123456789abcdef")
    "012345-6789ab-cdef"

    .. _RFC6920: https://www.ietf.org/rfc/rfc6920
    """
    segmented = []
    while h:
        segmented.append(h[:grouping])
        h = h[grouping:]
    return "-".join(segmented)

def _nih_checkdigit(h):
    """Luhn mod N algorithm in base 16 (hex) according to RFC6920_

    .. _RFC6920: https://www.ietf.org/rfc/rfc6920
    """
    ## Adopted from https://en.wikipedia.org/wiki/Luhn_mod_N_algorithm 
    ## pseudocode
    factor = 2
    total = 0
    base = 16
    digits = len(h)        
    # 0 if digits has even length, 1 if odd
    # (as we start doubling with the very last digit)
    parity = digits % 2
    for x in range(digits):
        digit = int(h[x], 16)            
        if x % 2 != parity:
            # double every second digit
            digit *= 2
            # slight less efficient, but more verbose:
            # if > 16:
            #   total += digit - 16 + 1
            # else: 
            #   total + digit
            total += sum(divmod(digit, 16))
        else:
            # Not doubled, must be <16
            total += digit
    # checkdigit that needs to be added to total
    # to get 0 after modulus
    remainder = (16-total) % 16
    # Return as hex digit
    return "%x" % remainder
